"""Meltano Setting Definitions."""

from __future__ import annotations

import ast
import json
import typing as t
from collections.abc import Mapping, Sequence
from datetime import date, datetime
from enum import Enum
from functools import cached_property

from meltano.core import utils
from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.error import Error

if t.TYPE_CHECKING:
    from ruamel.yaml import Node, Representer, ScalarNode

VALUE_PROCESSORS = {
    "nest_object": utils.nest_object,
    "upcase_string": lambda vlu: vlu.upper(),
    "stringify": lambda vlu: vlu if isinstance(vlu, str) else json.dumps(vlu),
}


class EnvVar:
    """Environment Variable class."""

    def __init__(self, definition: str) -> None:
        """Instantiate new EnvVar.

        Args:
            definition: Env var definition.
        """
        key = definition
        negated = False

        if definition.startswith("!"):
            key = definition[1:]
            negated = True

        self.key = key
        self.negated = negated

    @property
    def definition(self) -> str:
        """Return env var definition.

        Returns:
            Env var definition.
        """
        prefix = "!" if self.negated else ""
        return f"{prefix}{self.key}"

    def get(self, env) -> str:
        """Get env value.

        Args:
            env: Env to get value for.

        Returns:
            Env var value for given env var.
        """
        value = env[self.key]
        return str(not utils.truthy(value)) if self.negated else value


class SettingMissingError(Error):
    """A setting is missing."""

    def __init__(self, name: str) -> None:
        """Instantiate SettingMissingError.

        Args:
            name: Name of missing setting.
        """
        super().__init__(f"Cannot find setting {name}")


class YAMLEnum(str, Enum):
    """Serializable Enum class."""

    def __str__(self) -> str:
        """Return as string.

        Returns:
            This enum in string form.
        """
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj) -> str:
        """Represent as yaml.

        Args:
            dumper: YAML dumper.
            obj: Object to dump.

        Returns:
            Object in yaml string form.
        """
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))

    @classmethod
    def to_yaml(cls, representer: Representer, node: t.Any) -> ScalarNode:
        """Represent as yaml.

        Args:
            representer: YAML representer.
            node: Object to dump.

        Returns:
            Object in yaml string form.
        """
        return representer.represent_scalar("tag:yaml.org,2002:str", str(node))

    @classmethod
    def from_yaml(
        cls,
        constructor,  # noqa: ARG003
        node: Node,
    ) -> YAMLEnum:
        """Construct from yaml.

        Args:
            constructor: Class constructor.
            node: YAML node.

        Returns:
            Object from yaml node.
        """
        return cls(node.value)


class SettingKind(YAMLEnum):
    """Supported setting kinds."""

    STRING = "string"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DATE_ISO8601 = "date_iso8601"
    EMAIL = "email"
    PASSWORD = "password"  # noqa: S105
    OAUTH = "oauth"
    OPTIONS = "options"
    FILE = "file"
    ARRAY = "array"
    OBJECT = "object"
    HIDDEN = "hidden"

    @cached_property
    def is_sensitive(self) -> bool:
        """Return whether the setting kind is sensitive.

        Returns:
            True if the setting kind is sensitive.
        """
        return self in {
            SettingKind.PASSWORD,
            SettingKind.OAUTH,
        }


ParseValueExpectedType = t.TypeVar("ParseValueExpectedType")


class SettingDefinition(NameEq, Canonical):
    """Meltano SettingDefinition class."""

    name: str
    kind: SettingKind | None
    hidden: bool
    sensitive: bool
    _custom: bool

    def __init__(
        self,
        name: str | None = None,
        aliases: list[str] | None = None,
        env: str | None = None,
        env_aliases: list[str] | None = None,
        kind: str | None = None,
        value=None,
        label: str | None = None,
        documentation: str | None = None,
        description: str | None = None,
        tooltip: str | None = None,
        options: list | None = None,
        oauth: dict | None = None,
        placeholder: str | None = None,
        env_specific: bool | None = None,
        hidden: bool | None = None,
        sensitive: bool | None = None,
        custom: bool = False,
        value_processor=None,
        value_post_processor=None,
        **attrs,
    ):
        """Instantiate new SettingDefinition.

        Args:
            name: Setting name.
            aliases: Setting alias names.
            env: Setting target environment variable.
            env_aliases: Deprecated. Used to delegate alternative environment
                variables for overriding this setting's value.
            kind: Setting kind.
            value: Setting value.
            label: Setting label.
            documentation: Setting docs url.
            description: Setting description.
            tooltip: A phrase to provide additional information on this setting.
            options: Setting options.
            oauth: Setting OAuth provider details.
            placeholder: A placeholder value for this setting.
            env_specific: Flag for environment-specific setting.
            hidden: Hidden setting.
            sensitive: Sensitive setting.
            custom: Custom setting flag.
            value_processor: Used with `kind: object` to pre-process the keys
                in a particular way.
            value_post_processor: Used with `kind: object` to post-process the
                keys in a particular way.
            attrs: Keyword arguments to pass to parent class.
        """
        aliases = aliases or []
        env_aliases = env_aliases or []
        options = options or []
        oauth = oauth or {}

        kind = SettingKind(kind) if kind else None

        # Handle deprecated SettingKind.HIDDEN
        if kind is SettingKind.HIDDEN:
            # Override kind if hidden flag is set
            if hidden:
                kind = SettingKind.STRING

            # Prioritize kind over flag otherwise
            hidden = True

        # Handle deprecated SettingKind.PASSWORD and SettingKind.OAUTH
        if kind and kind.is_sensitive:
            # Override kind if sensitive flag is set
            if sensitive:
                kind = SettingKind.STRING

            # Prioritize kind over flag otherwise
            sensitive = True

        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            name=name,
            aliases=aliases,
            env=env,
            env_aliases=env_aliases,
            kind=kind,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            env_specific=env_specific,
            hidden=hidden,
            sensitive=sensitive,
            value_processor=value_processor,
            value_post_processor=value_post_processor,
            _custom=custom,
            **attrs,
        )

        self._verbatim.add("value")

    def __repr__(self) -> str:
        """Return string representation.

        Returns:
            String representation of this setting.
        """
        return f"<SettingDefinition {self.name} ({self.kind})>"

    @classmethod
    def from_missing(
        cls,
        defs: t.Iterable[SettingDefinition],
        config: dict,
        **kwargs,
    ) -> list[SettingDefinition]:
        """Create SettingDefinition instances for missing settings.

        Args:
            defs: Know setting definitions.
            config: Config dict.
            kwargs: Keyword arguments to pass to new SettingDefinition instances.

        Returns:
            A list of created custom SettingDefinition instances.
        """
        flat_config = utils.flatten(config, "dot")

        names = {setting.name for setting in defs}

        # Create custom setting definitions for unknown keys
        return [
            SettingDefinition.from_key_value(key, value, **kwargs)
            for key, value in flat_config.items()
            if key not in names
        ]

    @classmethod
    def from_key_value(
        cls,
        key: str,
        value: t.Any,
        custom: bool = True,
        default: t.Any | bool = False,
    ) -> SettingDefinition:
        """Create SettingDefinition instance from key-value pair.

        Args:
            key: Key.
            value: Value.
            custom: Custom setting flag.
            default: Default setting value.

        Returns:
            A SettingDefinition instance.
        """
        kind = None
        if isinstance(value, bool):
            kind = SettingKind.BOOLEAN
        elif isinstance(value, int):
            kind = SettingKind.INTEGER
        elif isinstance(value, dict):
            kind = SettingKind.OBJECT
        elif isinstance(value, list):
            kind = SettingKind.ARRAY

        attrs = {
            "name": key,
            "kind": kind,
            "custom": custom,
            "value": value if default else None,
        }

        return cls(**attrs)

    @property
    def is_extra(self) -> bool:
        """Return whether setting is a config extra.

        See https://docs.meltano.com/reference/command-line-interface#how-to-use-plugin-extras

        Returns:
            True if setting is a config extra.
        """  # noqa: E501
        return self.name.startswith("_")

    @property
    def is_custom(self) -> bool:
        """Return whether the setting is custom, i.e. user-defined in `meltano.yml`.

        Returns:
            True if the setting is custom (user defined).
        """
        return self._custom

    @property
    def is_redacted(self) -> bool:
        """Return whether the setting value is redacted.

        Returns:
            True if setting value is redacted.
        """
        return self.sensitive

    def env_vars(
        self,
        prefixes: list[str],
        include_custom: bool = True,
        for_writing: bool = False,
    ) -> list[EnvVar]:
        """Return environment variables with the provided prefixes.

        Args:
            prefixes: Env var prefixes to prepend.
            include_custom: Include custom env vars from `env_aliases`.
            for_writing: Include target env var from `env`.

        Returns:
            A list of EnvVar instances for this setting definition.
        """
        env_keys = []

        if self.env and for_writing:
            # this ensures we only write to specified `env:`
            env_keys.append(self.env)

        env_keys.extend(utils.to_env_var(prefix, self.name) for prefix in prefixes)
        if not for_writing:
            # read from setting name aliases
            for alias in self.aliases:
                env_keys.extend(utils.to_env_var(prefix, alias) for prefix in prefixes)

        if include_custom:
            env_keys.extend(self.env_aliases)

        return [EnvVar(key) for key in utils.uniques_in(env_keys)]

    @staticmethod
    def _parse_value(
        unparsed: str,
        expected_type_name: str,
        expected_type: type[ParseValueExpectedType],
    ) -> ParseValueExpectedType:
        """Parse a JSON string.

        Parsing is attempted first with `json.loads`, and then with
        `ast.literal_eval` as a fallback. It is used as a fallback because it
        correctly parses most inputs that `json.loads` can parse, but is more
        liberal about what it accepts. For example, `json.loads` requires
        double quotes for strings, but `ast.literal_eval` can use either single
        or double quotes.

        Args:
            unparsed: The JSON string.
            expected_type_name: The name of the expected type, e.g. "array".
                Used in the error message if parsing fails or the type is not
                as expected.
            expected_type: The Python type class of the expected type. Used to
                ensure that the parsed value is of the expected type.

        Raises:
            parse_error: Parsing failed, or the parsed value had an unexpected type.

        Returns:
            The parsed value.
        """
        parse_error = ValueError(
            f"Failed to parse JSON {expected_type_name} from string: {unparsed!r}",
        )
        try:
            parsed = json.loads(unparsed)
        except json.JSONDecodeError:
            try:  # noqa: WPS505
                parsed = ast.literal_eval(unparsed)
            except (
                ValueError,
                TypeError,
                SyntaxError,
                MemoryError,
                RecursionError,
            ) as ex:
                raise parse_error from ex
        if not isinstance(parsed, expected_type):
            raise parse_error
        return parsed

    def cast_value(self, value: t.Any) -> t.Any:  # noqa: C901
        """Cast given value.

        Args:
            value: Value to cast.

        Returns:
            Value cast according to specified setting definition kind.

        Raises:
            ValueError: If value is not of the expected type.
        """
        value = value.isoformat() if isinstance(value, (date, datetime)) else value

        if isinstance(value, str):
            if self.kind == SettingKind.BOOLEAN:
                return utils.truthy(value)
            if self.kind == SettingKind.INTEGER:
                return int(value)
            if self.kind == SettingKind.OBJECT:
                value = dict(
                    self._parse_value(value, "object", Mapping),  # type: ignore
                )
            elif self.kind == SettingKind.ARRAY:
                value = list(
                    self._parse_value(value, "array", Sequence),  # type: ignore
                )

        if (
            value is not None
            and self.kind == SettingKind.OPTIONS
            and all(opt["value"] != value for opt in self.options)
        ):
            error_message = f"'{value}' is not a valid choice for '{self.name}'"
            raise ValueError(error_message)

        processor = self.value_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def post_process_value(self, value: t.Any) -> t.Any:
        """Post-process given value.

        Args:
            value: Value to post-process.

        Returns:
            Value post-processed according to any post-processors specified for this
            setting definition.
        """
        processor = self.value_post_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def stringify_value(self, value: t.Any) -> str:
        """Return value in string form.

        Args:
            value: Value to stringify

        Returns:
            String form of the passed value.
        """
        if isinstance(value, str):
            return value

        if not self.kind or self.kind == SettingKind.STRING:
            return str(value)

        return json.dumps(value)



from __future__ import annotations

import logging
import typing as t
import warnings
from contextlib import closing

import pytest
from sqlalchemy import MetaData, create_engine
from sqlalchemy.exc import SAWarning
from sqlalchemy.orm import close_all_sessions, sessionmaker
from sqlalchemy.pool import NullPool

if t.TYPE_CHECKING:
    from meltano.core.project import Project


@pytest.fixture(scope="session", autouse=True)
def engine_uri_env(engine_uri: str) -> t.Generator:
    """Use the correct meltano database URI for these tests."""
    # No session monkey patch yet https://github.com/pytest-dev/pytest/issues/363
    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setenv("MELTANO_DATABASE_URI", engine_uri)
    try:
        yield
    finally:
        monkeypatch.undo()


@pytest.fixture()
def un_engine_uri(monkeypatch):
    """When we want to test functionality that doesn't use the current DB URI.

    Note that this fixture must run before the project fixture as
    src/meltano/core/db.py _engines has the engine cached.
    """
    monkeypatch.delenv("MELTANO_DATABASE_URI")


@pytest.fixture(scope="class", autouse=True)
def vacuum_db(engine_sessionmaker):
    try:
        yield
    finally:
        logging.debug("Cleaning system database...")
        engine, _ = engine_sessionmaker
        close_all_sessions()
        metadata = MetaData()
        metadata.reflect(bind=engine)
        metadata.drop_all(bind=engine)


@pytest.fixture(scope="class")
def engine_sessionmaker(engine_uri):
    engine = create_engine(engine_uri, poolclass=NullPool, future=True)
    return (engine, sessionmaker(bind=engine, future=True))


@pytest.fixture()
def connection(engine_sessionmaker):  # noqa: WPS442
    engine, _ = engine_sessionmaker
    connection = engine.connect()  # noqa: WPS442
    transaction = connection.begin()

    try:
        yield connection
    finally:
        with warnings.catch_warnings():
            # Ignore warnings about rolling back the same transaction twice
            warnings.filterwarnings(
                "ignore",
                "transaction already deassociated from connection",
                SAWarning,
            )
            transaction.rollback()
        connection.close()


@pytest.fixture()
def session(
    project: Project,  # noqa: ARG001
    engine_sessionmaker,
    connection,
):  # noqa: WPS442
    """Create a new database session for a test.

    Args:
        project: The `project` fixture.
        engine_sessionmaker: The `engine_sessionmaker` fixture.
        connection: The `connection` fixture.

    Yields:
        An ORM DB session for the given project, bound to the given connection.
    """
    _, create_session = engine_sessionmaker
    with closing(create_session(bind=connection)) as fixture_session:
        yield fixture_session



from __future__ import annotations

import platform
import sys
import typing as t
from importlib.util import module_from_spec, spec_from_file_location

import pytest
from mock import AsyncMock, mock

from meltano.core.plugin import PluginType
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_invoker import asyncio

if t.TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from meltano.core.plugin.superset import SupersetInvoker


def load_module_from_path(name: str, path: Path) -> ModuleType:
    """Load a module given its name and filesystem path.

    Replacement for the deprecated `imp.load_source`.

    Args:
        name: The name of the module as it would be in `sys.modules`.
        path: The path of the `.py` file.

    Returns:
        The imported module.
    """
    spec = spec_from_file_location(name, path)
    module = module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)
    return module


class TestSuperset:
    @pytest.fixture(scope="class")
    def subject(self, project_add_service):
        with mock.patch.object(PluginInstallService, "install_plugin"):
            return project_add_service.add(PluginType.UTILITIES, "superset")

    @pytest.mark.asyncio()  # noqa:  WPS210
    async def test_hooks(  # noqa:  WPS210
        self,
        subject,
        project,
        session,
        plugin_invoker_factory,
        monkeypatch,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        run_dir = project.run_dir("superset")
        config_path = run_dir.joinpath("superset_config.py")

        handle_mock = mock.Mock()
        handle_mock.name = subject.name
        handle_mock.wait = AsyncMock(return_value=0)
        handle_mock.returncode = 0

        original_exec = asyncio.create_subprocess_exec

        def popen_mock(cmd, *popen_args, **kwargs):
            assert kwargs["env"]["SUPERSET_HOME"] == str(run_dir)
            assert kwargs["env"]["SUPERSET_CONFIG_PATH"] == str(config_path)

            # first time, it creates the `superset.db`
            if {"db", "upgrade"}.issubset(popen_args):
                project.plugin_dir(subject, "superset.db").touch()
            # second time, it inits
            elif "init" in popen_args:  # noqa: SIM114
                return handle_mock
            # third time, it runs the requested command
            elif "--version" in popen_args:
                return handle_mock
            else:
                return original_exec(cmd, *popen_args, **kwargs)
            return handle_mock

        with mock.patch.object(
            asyncio,
            "create_subprocess_exec",
            side_effect=popen_mock,
        ) as popen, mock.patch(
            "meltano.core.plugin_invoker.PluginConfigService.configure",
        ):
            invoker: SupersetInvoker = plugin_invoker_factory(subject)
            # This ends up calling the hooks
            async with invoker.prepared(session):
                await invoker.invoke_async("--version")

                commands = [
                    popen_args
                    for _, popen_args, kwargs in popen.mock_calls
                    if popen_args and isinstance(popen_args, tuple)
                ]
                assert commands[0][1] == "db"
                assert commands[0][2] == "upgrade"
                assert commands[1][1] == "init"
                assert commands[2][1] == "--version"

                assert config_path.exists()
                assert project.plugin_dir(subject, "superset.db").exists()

                config_module = load_module_from_path("superset_config", config_path)

                config_keys = dir(config_module)  # noqa: WPS421
                assert "SQLALCHEMY_DATABASE_URI" in config_keys
                assert (
                    f'sqlite:///{project.plugin_dir(subject, "superset.db")}'
                    == config_module.SQLALCHEMY_DATABASE_URI
                )
                assert "SECRET_KEY" in config_keys

            # Test custom setting
            invoker.settings_service.set("SUPERSET_WEBSERVER_PORT", 5000)

            # Test config_path extra
            custom_config_filename = "custom_superset_config.py"
            custom_config_path = project.root.joinpath(custom_config_filename)
            custom_config_path.write_text('FOO_FAKE_SETTING = "fake_value"')
            monkeypatch.setitem(
                invoker.settings_service.config_override,
                "_config_path",
                custom_config_filename,
            )

            async with invoker.prepared(session):
                await invoker.invoke_async("--version")

                config_module = load_module_from_path("superset_config", config_path)

                config_keys = dir(config_module)  # noqa: WPS421
                # Verify default Meltano-managed settings are here
                assert "SQLALCHEMY_DATABASE_URI" in config_keys
                assert "SECRET_KEY" in config_keys
                # Verify custom Meltano-managed settings are here
                assert "SUPERSET_WEBSERVER_PORT" in config_keys
                assert config_module.SUPERSET_WEBSERVER_PORT == 5000
                # Verify settings from the custom config path are here
                assert "FOO_FAKE_SETTING" in config_keys
                assert config_module.FOO_FAKE_SETTING == "fake_value"

            assert not run_dir.joinpath("superset_config.py").exists()

    @pytest.mark.asyncio()
    async def test_before_cleanup(self, subject, plugin_invoker_factory):
        invoker: SupersetInvoker = plugin_invoker_factory(subject)

        assert not invoker.files["config"].exists()

        # No exception should be raised even though the file doesn't exist
        await invoker.cleanup()



from __future__ import annotations

import typing as t

import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    from fixtures.cli import MeltanoCliRunner
    from fixtures.docker import SnowplowMicro


@pytest.mark.parametrize(
    ("cmd", "expected"),
    (
        ("test", hash_sha256("dev")),
        ("--no-environment test", None),
        ("schedule list", None),
        ("--environment=dev schedule list", hash_sha256("dev")),
        ("--environment=prod schedule list", hash_sha256("prod")),
    ),
    ids=(
        "default-dev",
        "explicit-no-env",
        "default-no-env",
        "explicit-dev",
        "explicit-prod",
    ),
)
@pytest.mark.usefixtures("project")
def test_environment_name_hash(
    cmd: str,
    expected: str,
    snowplow: SnowplowMicro,
    cli_runner: MeltanoCliRunner,
):
    results = cli_runner.invoke(cli, cmd.split())
    assert_cli_runner(results)
    for event in snowplow.good():
        project_context = next(
            ctx
            for ctx in event["event"]["contexts"]["data"]
            if "project_context" in ctx["schema"]
        )
        assert project_context["data"]["environment_name_hash"] == expected



"""Service to manage meltano.yml."""

from __future__ import annotations

import logging
import os
import typing as t
from contextlib import contextmanager
from functools import cached_property

import yaml

from meltano.core import bundle
from meltano.core.setting_definition import SettingDefinition

if t.TYPE_CHECKING:
    from meltano.core.meltano_file import MeltanoFile
    from meltano.core.project import Project

logger = logging.getLogger(__name__)


class ConfigService:
    """Service to manage meltano.yml."""

    def __init__(self, project: Project):
        """Create a new project configuration service.

        Args:
            project: the project to configure.
        """
        self.project = project

    @cached_property
    def settings(self) -> list[SettingDefinition]:
        """Return the project settings.

        Returns:
            The project settings.
        """
        with open(str(bundle.root / "settings.yml")) as settings_yaml:
            settings_yaml_content = yaml.safe_load(settings_yaml)
        return [SettingDefinition.parse(x) for x in settings_yaml_content["settings"]]

    @cached_property
    def current_meltano_yml(self) -> MeltanoFile:
        """Return the current `meltano.yml` contents.

        Returns:
            The contents of `meltano.yml`.
        """
        return self.project.meltano

    @contextmanager
    def update_meltano_yml(self):
        """Update meltano.yml.

        This method is a context manager that will update meltano.yml with the
        contents of the context.

        Yields:
            The contents of meltano.yml.
        """
        with self.project.meltano_update() as meltano_yml:
            yield meltano_yml

    @contextmanager
    def update_active_environment(self):
        """Update active environment.

        Yields:
            active environment
        """
        environment = self.project.environment
        with self.update_meltano_yml() as meltano_yml:
            environments = meltano_yml.environments
            # find the proper environment to update
            env_idx, _ = next(
                (idx, env) for idx, env in enumerate(environments) if env == environment
            )
            yield environments[env_idx]

    @property
    def current_config(self):
        """Return the current configuration.

        Returns:
            The current configuration.
        """
        return self.current_meltano_yml.extras

    def update_config(self, config):
        """Update top-level Meltano configuration.

        Args:
            config: configuration dict
        """
        with self.update_meltano_yml() as meltano_yml:
            meltano_yml.extras = config

    def make_meltano_secret_dir(self):
        """Create the secret directory."""
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    @property
    def env(self):
        """Return the top-level env vars from meltano.yml.

        Returns:
            A dictionary of (unexpanded) environment variables.
        """
        return self.current_meltano_yml.env



"""Settings manager for Meltano plugins."""

from __future__ import annotations

import typing as t
from functools import cached_property

from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.settings_service import FeatureFlags, SettingsService
from meltano.core.utils import EnvVarMissingBehavior, expand_env_vars

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.setting_definition import SettingDefinition


class PluginSettingsService(SettingsService):  # noqa: WPS214
    """Settings manager for Meltano plugins."""

    def __init__(  # noqa: WPS210
        self,
        project: Project,
        plugin: ProjectPlugin,
        *args,
        **kwargs,
    ):
        """Create a new plugin settings manager.

        Args:
            project: The Meltano project.
            plugin: The Meltano plugin.
            args: Positional arguments to pass to the superclass.
            kwargs: Keyword arguments to pass to the superclass.
        """
        super().__init__(project, *args, **kwargs)
        self.plugin = plugin

        if self.project.environment:
            environment = self.project.environment
            self.environment_plugin_config = environment.get_plugin_config(
                self.plugin.type,
                self.plugin.name,
            )
        else:
            self.environment_plugin_config = None

        self.env_override = {
            # project level environment variables:
            **self.project.settings.env,
            # project level settings as env vars (e.g. `MELTANO_PROJECT_ID`):
            **self.project.settings.as_env(),
            # plugin level overrides, passed in as `**kwargs` and set to
            # `self.env_overrides` by `super().__init__` above:
            **self.env_override,
            # generated generic plugin settings as env vars
            # (e.g. `MELTANO_EXTRACT_NAME`):
            **self.plugin.info_env,
            # env vars stored under the `env:` key of the plugin definition:
            **self.plugin.env,
        }

        environment_env = {}
        if self.project.environment:
            with self.project.settings.feature_flag(
                FeatureFlags.STRICT_ENV_VAR_MODE,
                raise_error=False,
            ) as strict_env_var_mode:
                environment_env = {
                    var: expand_env_vars(
                        value,
                        self.env_override,
                        if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                    )
                    for var, value in self.project.environment.env.items()
                }
                # expand state_id_suffix
                self.project.environment.state_id_suffix = expand_env_vars(
                    self.project.environment.state_id_suffix,
                    {
                        **self.project.dotenv_env,
                        **self.env_override,
                    },
                    if_missing=EnvVarMissingBehavior(strict_env_var_mode),
                )

            # active Meltano Environment top level `env:` key
            self.env_override.update(environment_env)

        environment_plugin_env = (
            self.environment_plugin_config.env if self.environment_plugin_config else {}
        )
        # env vars stored under the `env:` key of the plugin definition of the
        # active meltano Environment
        self.env_override.update(environment_plugin_env)

    @property
    def project_settings_service(self):
        """Get the settings service for the active project.

        Returns:
            A project settings service for the active project.
        """
        return self.project.settings

    @property
    def label(self):
        """Get the label for this plugin.

        Returns:
            The label for this plugin.
        """
        return f"{self.plugin.type.descriptor} '{self.plugin.name}'"  # noqa: WPS237

    @property
    def docs_url(self):
        """Get the documentation URL for this plugin.

        Returns:
            The documentation URL for this plugin.
        """
        return self.plugin.docs

    def setting_env_vars(self, setting_def: SettingDefinition, for_writing=False):
        """Get environment variables for a setting.

        Args:
            setting_def: The setting definition.
            for_writing: Whether to get environment variables for writing.

        Returns:
            Environment variables for a setting.
        """
        return setting_def.env_vars(
            prefixes=self.plugin.env_prefixes(for_writing=for_writing),
            include_custom=self.plugin.is_shadowing or for_writing,
            for_writing=for_writing,
        )

    @property
    def db_namespace(self):
        """Return namespace for setting value records in system database.

        Returns:
            Namespace for setting value records in system database.
        """
        # "default" is included for legacy reasons
        return ".".join((self.plugin.type, self.plugin.name, "default"))

    @property
    def setting_definitions(self) -> list[SettingDefinition]:
        """Return definitions of supported settings.

        Returns:
            A list of setting definitions.
        """
        settings = self.plugin.settings_with_extras

        if self.environment_plugin_config is not None:
            settings.extend(
                self.environment_plugin_config.get_orphan_settings(settings),
            )

        return settings

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`.

        Returns:
            Current configuration in `meltano.yml`.
        """
        return self.plugin.config_with_extras

    @property
    def environment_config(self):
        """Return current environment configuration in `meltano.yml`.

        Returns:
            Current environment configuration in `meltano.yml`.
        """
        if self.environment_plugin_config:
            return self.environment_plugin_config.config_with_extras
        return {}

    def update_meltano_yml_config(self, config_with_extras):
        """Update configuration in `meltano.yml`.

        Args:
            config_with_extras: Configuration to update.
        """
        self.plugin.config_with_extras = config_with_extras
        self.project.plugins.update_plugin(self.plugin)

    def update_meltano_environment_config(self, config_with_extras: dict[str, t.Any]):
        """Update environment configuration in `meltano.yml`.

        Args:
            config_with_extras: Configuration to update.
        """
        self.environment_plugin_config.config_with_extras = config_with_extras
        self.project.plugins.update_environment_plugin(self.environment_plugin_config)

    @cached_property
    def inherited_settings_service(self):
        """Return settings service to inherit configuration from.

        Returns:
            Settings service to inherit configuration from.
        """
        return (
            type(self)(
                self.project,
                self.plugin.parent,
                env_override=self.env_override,
            )
            if isinstance(self.plugin.parent, ProjectPlugin)
            else None
        )

    def process_config(self, config):
        """Process configuration dictionary to be passed to plugin.

        Args:
            config: Configuration dictionary to process.

        Returns:
            Processed configuration dictionary.
        """
        return self.plugin.process_config(config)



"""Container specification."""

from __future__ import annotations

import shlex
import typing as t
from collections import defaultdict

from meltano.core.behavior.canonical import Canonical
from meltano.core.utils import expand_env_vars


def env_mapping_to_docker(mapping: dict[str, t.Any]) -> list[str]:
    """Convert a dictionary of environment variables to the <ENVVAR>=<VALUE> form.

    Args:
        mapping: Dictionary of environment variables.

    Returns:
        A list of <ENVVAR>=<VALUE> expression.

    >>> env_mapping_to_docker({"MY_VAR": "my_value", "DBT_TARGET": "postgres"})
    ['MY_VAR=my_value', 'DBT_TARGET=postgres']
    """
    return [f"{var}={value}" for var, value in mapping.items()]


class ContainerSpec(Canonical):
    """Container model for commands."""

    def __init__(
        self,
        image: str,
        *,
        command: str | None = None,
        entrypoint: str | None = None,
        ports: dict[str, str] | None = None,
        volumes: list[str] | None = None,
        env: dict[str, t.Any] | None = None,
    ):
        """Create a new container specification.

        Args:
            image: Container image.
            command: Command to run the container with.
            entrypoint: Overwrite the default `ENTRYPOINT` of the image.
            ports: Bind these ports.
            volumes: Bind mount these volumes.
            env: Environment variables.
        """
        super().__init__()

        self.image = image
        self.command = command
        self.entrypoint = entrypoint
        self.ports = ports or {}
        self.volumes = volumes or []
        self.env = env or {}

    def get_docker_config(self, *, additional_env: dict | None = None) -> dict:
        """Build a container configuration dictionary.

        Args:
            additional_env: Extend container environment with this mapping.

        Returns:
            Dictionary with container config.
        """
        env = {}

        if additional_env:
            env.update(additional_env)

        env.update(self.env)
        env_config = env_mapping_to_docker(env)

        volumes = [expand_env_vars(bind, env) for bind in self.volumes]

        exposed_ports: dict[str, t.Any] = {}
        port_bindings = defaultdict(list)

        for host_port, container_port in self.ports.items():
            exposed_ports[container_port] = {}
            port_bindings[container_port].append(
                {
                    "HostPort": host_port,
                    "HostIP": "0.0.0.0",  # noqa: S104. Binding to all interfaces is OK.
                },
            )

        return {
            "Cmd": shlex.split(self.command) if self.command else None,
            "Image": self.image,
            "Entrypoint": self.entrypoint,
            "Env": env_config,
            "ExposedPorts": exposed_ports,
            "HostConfig": {
                "PortBindings": port_bindings,
                "Binds": volumes,
            },
        }



"""Click parameter helper decorators."""

from __future__ import annotations

import functools

import click

from meltano.cli.utils import CliError
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationError
from meltano.core.project_settings_service import ProjectSettingsService


def database_uri_option(func):
    """Database URI Click option decorator.

    args:
        func: The function to decorate.
    """

    @click.option("--database-uri", help="System database URI.")
    def decorate(*args, database_uri=None, **kwargs):
        if database_uri:
            ProjectSettingsService.config_override["database_uri"] = database_uri

        return func(*args, **kwargs)

    return functools.update_wrapper(decorate, func)


class pass_project:  # noqa: N801
    """Pass current project to decorated CLI command function."""

    __name__ = "project"

    def __init__(self, migrate=False):
        """Instantiate decorator.

        args:
            migrate: Flag to perform database migration before passing the project.
        """
        self.migrate = migrate

    def __call__(self, func):
        """Return decorated function.

        args:
            func: The function to decorate.
        """

        @database_uri_option
        def decorate(*args, **kwargs):
            ctx = click.get_current_context()

            project = ctx.obj["project"]
            if not project:
                raise CliError(
                    f"`{ctx.command_path}` must be run inside a Meltano project.\n"  # noqa: EM102
                    "Use `meltano init <project_directory>` to create one.",
                )

            # register the system database connection
            engine, _ = project_engine(project, default=True)

            if self.migrate:
                from meltano.core.migration_service import MigrationService

                try:
                    migration_service = MigrationService(engine)
                    migration_service.upgrade(silent=True)
                except MigrationError as err:
                    raise CliError(str(err)) from err

            func(project, *args, **kwargs)

        return functools.update_wrapper(decorate, func)



"""Manager for Meltano project Environments."""

from __future__ import annotations

import typing as t

from meltano.core.environment import Environment
from meltano.core.utils import find_named

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class EnvironmentAlreadyExistsError(Exception):
    """An environment already exists."""

    def __init__(self, environment: Environment):
        """Create a new exception.

        Args:
            environment: An Environment object.
        """
        self.environment = environment
        super().__init__(f"An Environment named '{environment.name}' already exists.")


class EnvironmentService:
    """Meltano Service to manage Environment instances."""

    def __init__(self, project: Project):
        """Create a new EnvironmentService for a Meltano Project.

        Args:
            project: A Meltano Project.
        """
        self.project = project

    def add(self, name) -> Environment:
        """Create a new Environment in `meltano.yml`.

        Args:
            name: Name for new Environment.

        Returns:
            The newly added Environment object.
        """
        return self.add_environment(Environment(name=name))

    def add_environment(self, environment: Environment) -> Environment:
        """Add an Environment object to `meltano.yml`.

        Args:
            environment: An environment to add.

        Raises:
            EnvironmentAlreadyExistsError: If an environment with the same name
                already exists.

        Returns:
            The newly added environment.
        """
        with self.project.meltano_update() as meltano:
            # guard if it already exists
            # matching only by name works thanks to the `NameEq` mixin
            if environment in meltano.environments:
                raise EnvironmentAlreadyExistsError(environment=environment)

            meltano.environments.append(environment)

        return environment

    def list_environments(self) -> list[Environment]:
        """Enumerate existing Environments.

        Returns:
            A list of Environment objects.
        """
        return self.project.meltano.environments.copy()

    def remove(self, name: str) -> str:
        """Remove an Environment by name.

        Args:
            name: Name of the Environment that should be removed.

        Returns:
            The name of the removed Environment.
        """
        with self.project.meltano_update() as meltano:
            environment = find_named(
                self.list_environments(),
                name,
                obj_type=Environment,
            )

            # find the schedules plugin config
            meltano.environments.remove(environment)

        return name



"""meltano tasks/jobs (not state related jobs)."""

from __future__ import annotations

from collections.abc import Iterable

import structlog
import yaml
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical

logger = structlog.getLogger(__name__)


TASKS_JSON_SCHEMA = {
    "oneOf": [
        {"type": "string"},
        {
            "type": "array",
            "items": {
                "oneOf": [
                    {"type": "string"},
                    {"type": "array", "items": {"type": "string"}},
                ],
            },
        },
    ],
}


class InvalidTasksError(Exception):
    """Raised when a yaml str cannot be parsed into a valid tasks list."""

    def __init__(self, name: str, message: str):
        """Initialize a InvalidTasksError.

        Args:
            name: Name of the TaskSet that was invalid.
            message: The message describing the error.
        """
        self.name = name
        super().__init__(f"Job '{name}' has invalid tasks. {message}")


def _flat_split(items):
    for el in items:
        if isinstance(el, Iterable) and not isinstance(el, str):
            yield from _flat_split(el)
        elif " " in el:
            yield from _flat_split(el.split(" "))
        else:
            yield el


class TaskSets(NameEq, Canonical):
    """A named entity that holds one or more tasks that can be executed by Meltano."""

    def __init__(self, name: str, tasks: list[str] | list[list[str]]):
        """Initialize a `TaskSets`.

        Args:
            name: The name of the job.
            tasks: The tasks that associated with this job.
        """
        super().__init__()

        self.name = name
        self.tasks = tasks

    def _as_args(self, preserve_top_level: bool = False) -> list[str] | list[list[str]]:
        """Get job tasks as CLI args.

        Args:
            preserve_top_level: Whether to preserve the defined top level task
                list to allow for fine-grained executions.

        Returns:
            The run arguments.
        """
        if not preserve_top_level:
            return list(_flat_split(self.tasks))

        flattened = []
        for task in self.tasks:
            if isinstance(task, str):
                flattened.append(task.split(" "))
            else:
                flattened.append(list(_flat_split(task)))
        return flattened

    @property
    def flat_args(self) -> list[str] | list[list[str]]:
        """Convert job's tasks to a single invocable representations.

        For passing as a cli argument or as block names.

        Example:
            >>> TaskSets(name="foo", tasks=["tap target", "some:cmd"]).flat_args
            ["tap", "target", "some:cmd"]

        Returns:
            The run arguments.
        """
        return self._as_args()

    @property
    def flat_args_per_set(self) -> list[str] | list[list[str]]:
        """Convert the job's tasks into perk task representations.

        Preserves top level list hierarchy.

        Example:
        >>> TaskSets(name="foo", tasks=[["a b"], ["some:cmd"]).flat_args_per_set
        [["a", "b"], ["some:cmd"]]

        Returns:
            The per-task run arguments.
        """  # noqa: F721
        return self._as_args(preserve_top_level=True)


def tasks_from_yaml_str(name: str, yaml_str: str) -> TaskSets:
    """Create a TaskSets from a yaml string.

    The resulting object is validated against the `TASKS_JSON_SCHEMA`.

    Args:
        name: The name of the job.
        yaml_str: The yaml string.

    Returns:
        The TaskSets.

    Raises:
        InvalidTasksError: If the yaml string failed to parse or failed to
            validate against the `TASKS_JSON_SCHEMA`.
    """
    tasks = []
    try:
        tasks = yaml.safe_load(yaml_str)
    except yaml.parser.ParserError as yerr:
        raise InvalidTasksError(
            name,
            f"Failed to parse yaml '{yaml_str}': {yerr}",
        ) from yerr

    try:
        validate(instance=tasks, schema=TASKS_JSON_SCHEMA)
    except ValidationError as verr:
        raise InvalidTasksError(
            name,
            f"Failed to validate task schema: {verr}",
        ) from verr

    # Handle the special case of a single task
    if isinstance(tasks, str):
        tasks = [tasks]

    return TaskSets(name=name, tasks=tasks)



from __future__ import annotations

import typing as t
from pathlib import Path

import yaml

from meltano.core import bundle
from meltano.core.plugin.file import FilePlugin

if t.TYPE_CHECKING:
    import os

    from meltano.core.project import Project


class MeltanoFilePlugin(FilePlugin):
    overwrite = {"meltano.yml"}

    def __init__(self):
        """Initialize the MeltanoFilePlugin."""
        super().__init__(None, None)

    def file_contents(
        self,
        project,  # noqa: ARG002
    ) -> dict[Path, str]:
        """Get a mapping of paths to contents.

        Args:
            project: The Meltano project:

        Returns:
            A mapping.
        """
        with (bundle.root / "initialize.yml").open() as f:
            return {
                Path(relative_path): content
                for relative_path, content in yaml.safe_load(f).items()
            }

    def update_config(self, project):  # noqa: ARG002, D102
        return {}

    def files_to_create(
        self,
        project: Project,
        paths_to_update: list[str],
    ) -> dict[Path, str]:
        """Return the contents of the files to be created in the project.

        Args:
            project: The Meltano project.
            paths_to_update: The paths of the files to be updated.

        Returns:
            A dictionary of file names and their contents.
        """
        return self.project_file_contents(project, paths_to_update)

    def project_file_contents(
        self,
        project: Project,
        paths_to_update: list[str],  # noqa: ARG002
    ) -> dict[Path, str]:
        """Return the contents of the files to be created or updated in the project.

        Args:
            project: The Meltano project.
            paths_to_update: The paths of the files to be updated.

        Returns:
            A dictionary of file names and their contents.
        """
        return self.file_contents(project)

    def write_file(
        self,
        project: Project,
        relative_path: os.PathLike,
        content: str,
    ) -> bool:
        """Write the file to the project.

        Args:
            project: The Meltano project.
            relative_path: The relative path of the file.
            content: The contents of the file.

        Returns:
            True if the file was written, False otherwise.
        """
        project_path = project.root_dir(relative_path)
        if project_path.exists() and not any(
            project_path.match(path) for path in self.overwrite
        ):
            return False

        project_path.parent.mkdir(parents=True, exist_ok=True)
        project_path.write_text(content)

        return True



from __future__ import annotations

import typing as t

from sqlalchemy import types
from sqlalchemy.orm import Mapped, mapped_column

from meltano.core.sqlalchemy import StrPK  # noqa: TCH001

from .models import SystemModel


class Setting(SystemModel):
    __tablename__ = "plugin_settings"

    # represent the mapping to the ENV
    label: Mapped[t.Optional[str]]  # noqa: UP007
    description: Mapped[t.Optional[str]] = mapped_column(types.Text)  # noqa: UP007

    # represent a materialized path to support
    # a nested configuration.
    name: Mapped[StrPK]
    namespace: Mapped[t.Optional[StrPK]]  # noqa: UP007
    value: Mapped[t.Optional[str]] = mapped_column(types.PickleType)  # noqa: UP007
    enabled: Mapped[bool] = mapped_column(default=False)

    def __repr__(self):
        enabled_marker = "E" if self.enabled else ""
        return f"<({self.namespace}) {self.name}={self.value} {enabled_marker}>"



from __future__ import annotations

import json

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli


class TestCliJob:
    @pytest.mark.usefixtures("session", "project")
    def test_job_add(self, cli_runner, task_sets_service):
        # singular task with job
        with mock.patch(
            "meltano.cli.job.TaskSetsService",
            return_value=task_sets_service,
        ), mock.patch("meltano.cli.job._validate_tasks", return_value=True):
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-mock",
                    "--tasks",
                    "'tap-mock target-mock'",
                ],
            )
            assert_cli_runner(res)
            task_sets = task_sets_service.get("job-mock")

            assert task_sets.name == "job-mock"
            assert task_sets.tasks == ["tap-mock target-mock"]

            # valid pseudo-list of tasks
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-mock2",
                    "--tasks",
                    "['tap-mock target-mock', 'tap-mock2 target-mock2']",
                ],
            )
            assert_cli_runner(res)

            task_sets = task_sets_service.get("job-mock2")
            assert task_sets.name == "job-mock2"
            assert task_sets.tasks == ["tap-mock target-mock", "tap-mock2 target-mock2"]

            # verify that you can't add a job with the same name
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-mock2",
                    "--tasks",
                    "['tap-mock target-mock', 'tap-mock2 target-mock2']",
                ],
            )
            assert res.exit_code == 1
            assert "Job 'job-mock2' already exists" in str(res.exception)

            # invalid task - schema validation fails
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-mock-bad-schema",
                    "--tasks",
                    '["tap-gitlab target-jsonl", "dbt:run", 5]',
                ],
                catch_exceptions=True,
            )
            assert res.exit_code == 1
            assert "Failed to validate task schema" in str(res.exception)

            # invalid task - yaml parsing failure
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-mock-bad-yaml",
                    "--tasks",
                    "['tap-mock target-mock'",
                ],
                catch_exceptions=True,
            )
            assert res.exit_code == 1
            assert "Failed to parse yaml" in str(res.exception)

    @pytest.mark.usefixtures("session", "project")
    def test_job_set(self, cli_runner, task_sets_service):
        # singular task with job
        with mock.patch(
            "meltano.cli.job.TaskSetsService",
            return_value=task_sets_service,
        ), mock.patch("meltano.cli.job._validate_tasks", return_value=True):
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-set-mock",
                    "--tasks",
                    "'tap-mock target-mock'",
                ],
            )
            assert_cli_runner(res)
            assert task_sets_service.exists("job-set-mock")

            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "set",
                    "job-set-mock",
                    "--tasks",
                    "'tap2-mock target2-mock'",
                ],
            )

            task_sets = task_sets_service.get("job-set-mock")
            assert task_sets.name == "job-set-mock"
            assert task_sets.tasks == ["tap2-mock target2-mock"]

    @pytest.mark.order(after="test_job_add")
    @pytest.mark.usefixtures("session", "project")
    def test_job_remove(self, cli_runner, task_sets_service):
        # singular task with job
        with mock.patch(
            "meltano.cli.job.TaskSetsService",
            return_value=task_sets_service,
        ), mock.patch("meltano.cli.job._validate_tasks", return_value=True):
            res = cli_runner.invoke(
                cli,
                [
                    "job",
                    "add",
                    "job-remove-mock",
                    "--tasks",
                    "'tap-mock target-mock'",
                ],
            )
            assert_cli_runner(res)
            assert task_sets_service.exists("job-mock")

            res = cli_runner.invoke(cli, ["job", "remove", "job-remove-mock"])
            assert_cli_runner(res)
            assert not task_sets_service.exists("job-remove-mock")

    @pytest.mark.usefixtures("session", "project")
    def test_job_list(self, cli_runner, task_sets_service):
        # singular task with job
        with mock.patch(
            "meltano.cli.job.TaskSetsService",
            return_value=task_sets_service,
        ), mock.patch("meltano.cli.job._validate_tasks", return_value=True):
            cli_args = [
                "job",
                "add",
                "job-list-mock",
                "--tasks",
                "'tap-mock target-mock'",
            ]
            res = cli_runner.invoke(cli, cli_args)
            assert_cli_runner(res)
            assert task_sets_service.exists("job-list-mock")

            cli_args = [
                "job",
                "add",
                "job-list-mock2",
                "--tasks",
                "'tap-mock2 target-mock2'",
            ]
            res = cli_runner.invoke(cli, cli_args)
            assert_cli_runner(res)
            assert task_sets_service.exists("job-list-mock2")

            # test vanilla
            res = cli_runner.invoke(cli, ["job", "list"])
            assert_cli_runner(res)
            assert "job-list-mock: ['tap-mock target-mock']" in res.output
            assert "job-list-mock2: ['tap-mock2 target-mock2']" in res.output

            # test singular json list
            res = cli_runner.invoke(
                cli,
                ["job", "list", "--format=json", "job-list-mock"],
            )
            assert_cli_runner(res)
            output = json.loads(res.output)
            assert output["job_name"] == "job-list-mock"
            assert output["tasks"] == ["tap-mock target-mock"]



"""A module for the UtilityPlugin type."""

from __future__ import annotations

from meltano.core.plugin import BasePlugin, PluginType


class UtilityPlugin(BasePlugin):
    """A plugin type for arbitrary pip executables."""

    __plugin_type__ = PluginType.UTILITIES



from __future__ import annotations

import json
import typing as t

import pytest

from meltano.core.plugin.base import BasePlugin, PluginDefinition, PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_lock_service import (
    LockfileAlreadyExistsError,
    PluginLock,
    PluginLockService,
)

if t.TYPE_CHECKING:
    from meltano.core.project import Project


@pytest.fixture()
def plugin():
    plugin_definition = PluginDefinition(
        PluginType.EXTRACTORS,
        name="tap-locked",
        namespace="tap_locked",
        foo="bar",
        variants=[
            {
                "name": "meltano",
                "pip_url": "meltano-tap-locked",
                "repo": "https://gitlab.com/meltano/tap-locked",
                "baz": "qux",
            },
            {
                "name": "singer-io",
                "original": True,
                "deprecated": True,
                "pip_url": "tap-locked",
                "repo": "https://github.com/singer-io/tap-locked",
            },
        ],
    )
    variant = plugin_definition.find_variant("meltano")
    base_plugin = BasePlugin(plugin_definition, variant)

    parent_plugin = ProjectPlugin(
        base_plugin.type,
        base_plugin.name,
        variant=variant,
    )
    parent_plugin.parent = base_plugin

    plugin = ProjectPlugin(
        base_plugin.type,
        f"{base_plugin.name}--two",
        inherit_from=parent_plugin.name,
    )
    plugin.parent = parent_plugin
    return plugin


class TestPluginLock:
    @pytest.fixture()
    def subject(self, project: Project, plugin: ProjectPlugin):
        return PluginLock(project, plugin)

    def test_path(self, subject: PluginLock):
        assert subject.path.parts[-3:] == (
            "plugins",
            "extractors",
            "tap-locked--meltano.lock",
        )

    @pytest.mark.order(before="test_load")
    def test_save(self, subject: PluginLock):
        assert not subject.path.exists()
        subject.save()
        assert subject.path.exists()

    def test_load(self, subject: PluginLock, plugin: ProjectPlugin):
        subject.save()
        loaded = subject.load()
        assert loaded.name == plugin.inherit_from
        assert loaded.variant == plugin.variant.name
        assert loaded.settings == plugin.settings


class TestPluginLockService:
    @pytest.fixture()
    def subject(self, project: Project):
        return PluginLockService(project)

    def test_save(
        self,
        subject: PluginLockService,
        project: Project,
        plugin: ProjectPlugin,
    ):
        plugin_lock = PluginLock(project, plugin)
        assert not plugin_lock.path.exists()

        subject.save(plugin)
        assert plugin_lock.path.exists()

        with plugin_lock.path.open() as lock_file:
            lock_json = json.load(lock_file)
            assert lock_json["foo"] == "bar"
            assert lock_json["baz"] == "qux"

        with pytest.raises(LockfileAlreadyExistsError) as exc_info:
            subject.save(plugin)

        assert exc_info.value.plugin == plugin



"""THIS IS A NO-OP.

IT EXISTS ONLY TO PREVENT USERS WHO APPLIED THIS MIGRATION
FROM ENDING UP WITH AN ORPHANED alembic_version IN THEIR DB.

Revision ID: f4c225a9492f
Revises: 5b43800443d1
Create Date: 2022-09-02 09:44:05.581824

"""
from __future__ import annotations

from alembic import op

# revision identifiers, used by Alembic.
revision = "f4c225a9492f"
down_revision = "5b43800443d1"
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""This module contains the SingerMapper class as well as a supporting methods."""

from __future__ import annotations

import json
import typing as t

import structlog

from meltano.core.behavior.hookable import hook
from meltano.core.setting_definition import SettingDefinition, SettingKind

from . import PluginType, SingerPlugin

if t.TYPE_CHECKING:
    from meltano.core.plugin_invoker import PluginInvoker

logger = structlog.stdlib.get_logger(__name__)


class SingerMapper(SingerPlugin):
    """A SingerMapper is a singer spec compliant stream mapper."""

    __plugin_type__ = PluginType.MAPPERS

    EXTRA_SETTINGS = [
        SettingDefinition(
            name="_mappings",
            kind=SettingKind.ARRAY,
            aliases=["mappings"],
            value={},
        ),
        SettingDefinition(
            name="_mapping_name",
            kind=SettingKind.STRING,
            aliases=["mapping_name"],
            value=None,
        ),
    ]

    def exec_args(self, plugin_invoker: PluginInvoker):
        """Return the arguments to be passed to the plugin's executable."""
        return ["--config", plugin_invoker.files["config"]]

    @property
    def config_files(self):
        """Return the configuration files required by the plugin."""
        return {"config": f"mapper.{self.instance_uuid}.config.json"}

    @hook("before_configure")
    async def before_configure(
        self,
        invoker: PluginInvoker,
        session,  # noqa: ARG002
    ):
        """Create configuration file."""  # noqa: DAR101
        config_path = invoker.files["config"]

        config_payload: dict = {}
        with open(config_path, "w") as config_file:
            config_payload = self._get_mapping_config(invoker.plugin.extra_config)
            json.dump(config_payload, config_file, indent=2)

        logger.debug(
            "Created configuration",
            config_path=config_path,
            plugin_name=invoker.plugin.name,
            mapping_name=invoker.plugin_config_extras["_mapping_name"],
            mapping_config=config_payload,
        )

    @staticmethod
    def _get_mapping_config(extra_config: dict) -> dict | None:
        for mapping in extra_config.get("_mappings", []):  # noqa: RET503
            if mapping.get("name") == extra_config.get("_mapping_name"):
                return mapping["config"]



"""Defines MeltanoInvoker."""

from __future__ import annotations

import os
import platform
import subprocess
import sys
import typing as t
from pathlib import Path

from meltano.core.project_settings_service import SettingValueStore
from meltano.core.tracking import Tracker

if t.TYPE_CHECKING:
    from collections.abc import Mapping

    from meltano.core.project import Project

MELTANO_COMMAND = "meltano"


class MeltanoInvoker:
    """Class used to find and invoke all commands passed to it."""

    def __init__(self, project: Project):
        """
        Load the class with the project and service settings.

        Args:
            project: Project instance.
        """
        self.project = project
        self.tracker = Tracker(project)

    def invoke(
        self,
        args: t.Iterable[str],
        command: str = MELTANO_COMMAND,
        env: dict[str, str] | None = None,
        **kwargs: t.Any,
    ) -> subprocess.CompletedProcess:
        """Invoke meltano or other provided command.

        Args:
            args: CLI arguments to pass to the command.
            command: Executable to invoke.
            env: Extra environment variables to use for the subprocess.
            kwargs: Keyword arguments for `subprocess.run`.

        Returns:
            A `CompletedProcess` class object from `subprocess.run`.
        """
        return subprocess.run(
            [self._executable_path(command), *args],
            **kwargs,
            env=self._executable_env(env),
        )

    def _executable_path(self, command):
        if command == MELTANO_COMMAND:
            # This symlink is created by Project.activate
            executable_symlink = self.project.run_dir().joinpath("bin")
            if executable_symlink.exists():
                return str(executable_symlink)

        executable = Path(
            os.path.dirname(sys.executable),
            f"{command}.exe" if platform.system() == "Windows" else command,
        )

        # Fall back on expecting command to be in the PATH
        return str(executable) if executable.exists() else command

    def _executable_env(self, env: Mapping[str, str] | None = None) -> dict[str, str]:
        if env is None:
            env = {}
        return {
            # Include env that project settings are evaluated in
            **self.project.settings.env,
            # Include env for settings explicitly overridden using CLI flags
            **self.project.settings.as_env(source=SettingValueStore.CONFIG_OVERRIDE),
            # Include explicitly provided env
            **env,
            # Include telemetry env vars
            **self.tracker.env,
        }



"""State management in CLI."""

from __future__ import annotations

import json
import re
import typing as t
from datetime import datetime as dt
from functools import partial, reduce
from operator import xor

import click
import structlog

from meltano.cli.params import pass_project
from meltano.cli.utils import CliEnvironmentBehavior, InstrumentedCmd, InstrumentedGroup
from meltano.core.block.parser import BlockParser
from meltano.core.db import project_engine
from meltano.core.job import Payload
from meltano.core.state_service import InvalidJobStateError, StateService

if t.TYPE_CHECKING:
    from meltano.core.project import Project

STATE_SERVICE_KEY = "state_service"

logger = structlog.getLogger(__name__)


class MutuallyExclusiveOptionsError(Exception):
    """Mutually exclusive options are provided incorrectly."""

    def __init__(self, *options: str) -> None:
        """Instantiate the error.

        Args:
            options: the mutually exclusive options that were incorrectly provided.
        """
        super().__init__(*options)
        self.options = options

    def __str__(self) -> str:
        """Represent the error as a string."""
        return f"Must provide exactly one of: {','.join(self.options)}"


def _prompt_for_confirmation(prompt):
    """Wrap destructive CLI commands which should prompt the user for confirmation."""

    def _prompt_callback(ctx: click.Context, param, value: bool):  # noqa: ARG001
        if not value and not click.confirm(prompt):
            ctx.exit(1)

    return click.option(
        "--force",
        is_flag=True,
        expose_value=False,
        callback=_prompt_callback,
        help="Don't prompt for confirmation.",
    )


prompt_for_confirmation = partial(
    _prompt_for_confirmation,
    prompt="This is a destructive command. Continue?",
)


def state_service_from_state_id(project: Project, state_id: str) -> StateService | None:
    """Instantiate by parsing a state_id."""
    state_id_re = re.compile(
        r"^(?P<env>.+):(?P<tap>.+)-to-(?P<target>.+?)(?:\:(?P<suffix>.+))?(?<=[^\:])$",
    )
    if match := state_id_re.match(state_id):
        # If the state_id matches convention (i.e., job has been run via "meltano run"),
        # try parsing into BlockSet.
        # This way, we get BlockSet validation and raise an error if no
        # plugin in the BlockSet has "state" capability
        try:
            if not project.environment:
                logger.warning(
                    "Running state operation for environment "  # noqa: G004
                    f"'{match['env']}' outside of an environment",
                )

            elif project.environment.name != match["env"]:
                logger.warning(
                    f"Environment '{match['env']}' used in state operation does "  # noqa: G004
                    f"not match current environment '{project.environment.name}'.",
                )

            project.activate_environment(match["env"])
            blocks = [match["tap"], match["target"]]
            parser = BlockParser(logger, project, blocks)
            return next(parser.find_blocks()).state_service
        except Exception:
            logger.warning("No plugins found for provided state_id.")
    # If provided state_id does not match convention (i.e., run via "meltano elt"),
    # use the standalone StateService in the CLI context.
    return None


@click.group(
    cls=InstrumentedGroup,
    name="state",
    short_help="Manage Singer state.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_ignore_default,
)
@click.pass_context
@pass_project(migrate=True)
def meltano_state(project: Project, ctx: click.Context):
    """
    Manage state.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#state
    """
    _, sessionmaker = project_engine(project)
    session = sessionmaker(future=True)
    ctx.obj[STATE_SERVICE_KEY] = StateService(project, session)  # noqa: WPS204


@meltano_state.command(cls=InstrumentedCmd, name="list")
@click.option("--pattern", type=str, help="Filter state IDs by pattern.")
@click.pass_context
def list_state(ctx: click.Context, pattern: str | None):  # noqa: WPS125
    """List all state_ids for this project.

    Optionally pass a glob-style pattern to filter state_ids by.
    """
    state_service: StateService = ctx.obj[STATE_SERVICE_KEY]
    if states := state_service.list_state(pattern):
        for state_id, state in states.items():
            if state:
                try:
                    state_service.validate_state(json.dumps(state))
                except (InvalidJobStateError, json.decoder.JSONDecodeError):
                    click.secho(state_id, fg="red")
                else:
                    click.secho(state_id, fg="green")
            else:
                click.secho(state_id, fg="yellow")
    else:
        logger.info("No state IDs found.")


@meltano_state.command(cls=InstrumentedCmd, name="copy")
@prompt_for_confirmation(
    prompt="This will overwrite state for the destination. Continue?",
)
@click.argument("src-state-id", type=str)
@click.argument("dst-state-id", type=str)
@pass_project(migrate=True)
@click.pass_context
def copy_state(
    ctx: click.Context,
    project: Project,
    src_state_id: str,
    dst_state_id: str,
):
    """Copy state to another job ID."""
    # Retrieve state for copying
    state_service: StateService = (
        state_service_from_state_id(project, src_state_id) or ctx.obj[STATE_SERVICE_KEY]
    )

    state_service.copy_state(src_state_id, dst_state_id)

    logger.info(
        f"State for {dst_state_id} was successfully copied from "  # noqa: G004
        f"{src_state_id} at {dt.utcnow():%Y-%m-%d %H:%M:%S}.",  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="move")
@prompt_for_confirmation(
    prompt=(
        "This will clear the source state and overwrite destination state. Continue?"
    ),
)
@click.argument("src-state-id", type=str)
@click.argument("dst-state-id", type=str)
@pass_project(migrate=True)
@click.pass_context
def move_state(
    ctx: click.Context,
    project: Project,
    src_state_id: str,
    dst_state_id: str,
):
    """Move state to another job ID, clearing the original."""
    # Retrieve state for moveing
    state_service: StateService = (
        state_service_from_state_id(project, dst_state_id) or ctx.obj[STATE_SERVICE_KEY]
    )

    state_service.move_state(src_state_id, dst_state_id)

    logger.info(
        f"State for {src_state_id} was successfully moved to {dst_state_id} "  # noqa: G004
        f"at {dt.utcnow():%Y-%m-%d %H:%M:%S}.",  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="merge")
@click.option(
    "--from-state-id",
    type=str,
    help="Merge state from an existing state ID.",
)
@click.option(
    "--input-file",
    type=click.Path(exists=True),
    help="Merge state from a JSON file containing Singer state.",
)
@click.argument("state-id", type=str)
@click.argument("state", type=str, required=False)
@pass_project(migrate=True)
@click.pass_context
def merge_state(
    ctx: click.Context,
    project: Project,
    state_id: str,
    state: str | None,
    input_file: click.Path | None,
    from_state_id: str | None,
):
    """Add bookmarks to existing state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    mutually_exclusive_options = {
        "--input-file": input_file,
        "STATE": state,
        "--from-state-id": from_state_id,
    }
    if not reduce(xor, (bool(x) for x in mutually_exclusive_options.values())):
        raise MutuallyExclusiveOptionsError(*mutually_exclusive_options)
    if input_file:
        with open(input_file) as state_f:
            state_service.add_state(
                state_id,
                state_f.read(),
                payload_flags=Payload.INCOMPLETE_STATE,
            )
    elif state:
        state_service.add_state(state_id, state, payload_flags=Payload.INCOMPLETE_STATE)
    elif from_state_id:
        state_service.merge_state(from_state_id, state_id)
    logger.info(
        f"State for {state_id} was successfully "  # noqa: G004
        f"merged at {dt.utcnow():%Y-%m-%d %H:%M:%S}.",  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="set")
@prompt_for_confirmation(
    prompt="This will overwrite the state's current value. Continue?",
)
@click.option(
    "--input-file",
    type=click.Path(exists=True),
    help="Set state from json file containing Singer state.",
)
@click.argument("state-id")
@click.argument("state", type=str, required=False)
@pass_project(migrate=True)
@click.pass_context
def set_state(
    ctx: click.Context,
    project: Project,
    state_id: str,
    state: str | None,
    input_file: click.Path | None,
):
    """Set state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    mutually_exclusive_options = {
        "--input-file": input_file,
        "STATE": state,
    }
    if not reduce(xor, (bool(x) for x in mutually_exclusive_options.values())):
        raise MutuallyExclusiveOptionsError(*mutually_exclusive_options)
    if input_file:
        with open(input_file) as state_f:
            state_service.set_state(state_id, state_f.read())
    elif state:
        state_service.set_state(state_id, state)
    logger.info(
        f"State for {state_id} was successfully set "  # noqa: G004
        f"at {dt.utcnow():%Y-%m-%d %H:%M:%S}.",  # noqa: WPS323
    )


@meltano_state.command(cls=InstrumentedCmd, name="get")  # noqa: WPS46
@click.argument("state-id")
@pass_project(migrate=True)
@click.pass_context
def get_state(ctx: click.Context, project: Project, state_id: str):  # noqa: WPS463
    """Get state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    retrieved_state = state_service.get_state(state_id)
    click.echo(json.dumps(retrieved_state))


@meltano_state.command(cls=InstrumentedCmd, name="clear")
@prompt_for_confirmation(prompt="This will clear state for the job. Continue?")
@click.argument("state-id")
@pass_project(migrate=True)
@click.pass_context
def clear_state(ctx: click.Context, project: Project, state_id: str):
    """Clear state."""
    state_service: StateService = (
        state_service_from_state_id(project, state_id) or ctx.obj[STATE_SERVICE_KEY]
    )
    state_service.clear_state(state_id)



from __future__ import annotations

import pytest

from meltano.core.task_sets import TaskSets
from meltano.core.task_sets_service import (
    JobAlreadyExistsError,
    JobNotFoundError,
    TaskSetsService,
)


@pytest.fixture(scope="session")
def create_task_set():
    def make(name):
        return TaskSets(name=name, tasks=["tap-mock target-mock"])

    return make


class TestTaskSetsService:
    @pytest.fixture()
    def subject(self, task_sets_service):
        return task_sets_service

    @pytest.mark.order(0)
    def test_add(self, subject: TaskSetsService, create_task_set):
        count = 10
        jobs = [create_task_set(f"test_job_{idx}") for idx in range(count)]

        for job in jobs:
            subject.add(job)

        assert subject.list() == jobs

        # verify that we can not add a job with the same name
        with pytest.raises(JobAlreadyExistsError):
            subject.add(jobs[0])

    def test_update(self, subject: TaskSetsService, create_task_set):
        job = subject.list()[0]
        job.tasks = ["tap-mock target-mock updated:addition"]
        subject.update(job)

        assert subject.get(job.name).tasks == ["tap-mock target-mock updated:addition"]

        nonexistent = create_task_set("does-not-exist")
        with pytest.raises(JobNotFoundError):
            subject.update(nonexistent)

    @pytest.mark.usefixtures("create_task_set")
    def test_remove(self, subject: TaskSetsService):
        jobs = subject.list()
        subject.remove(jobs[0].name)
        assert subject.list() == jobs[1:]
        assert subject.exists(jobs[0].name) is False

        # verify that we can not remove a job that does not exist
        with pytest.raises(JobNotFoundError):
            subject.remove(jobs[0].name)

    @pytest.mark.usefixtures("create_task_set")
    def test_get(self, subject: TaskSetsService):
        jobs = subject.list()

        assert subject.get(jobs[0].name) == jobs[0]

        # verify that we can not get a job that does not exist
        with pytest.raises(JobNotFoundError):
            subject.get("non-existent")

    @pytest.mark.usefixtures("create_task_set")
    def test_exists(self, subject: TaskSetsService):
        job = subject.list()[0]
        assert subject.exists(job.name)
        assert not subject.exists("non-existent")

    def test_list(self, subject: TaskSetsService, create_task_set):
        expected_jobs = [create_task_set(f"test_list_{idx}") for idx in range(3)]

        for job in expected_jobs:
            subject.add(job)

        result = subject.list()

        for job in expected_jobs:
            assert job in result



"""Defines JobState model class."""

from __future__ import annotations

import json
import typing as t
from datetime import datetime

from sqlalchemy.orm import Mapped, Session, mapped_column

from meltano.core.job import JobFinder, Payload
from meltano.core.models import SystemModel
from meltano.core.sqlalchemy import StateType  # noqa: TCH001
from meltano.core.utils import merge

if t.TYPE_CHECKING:
    from io import TextIOWrapper


SINGER_STATE_KEY = "singer_state"
STATE_ID_COMPONENT_DELIMITER = ":"


class JobState(SystemModel):  # noqa: WPS214
    """Model class that represents the current state of a given job.

    Modified during `meltano elt` or `meltano run` runs whenever a
    STATE message is emitted by a Singer target. Also written and read
    by `meltano state` CLI invocations. Only holds the _current_ state
    for a given state_id. Full job run history is held by the Job model.
    """

    __tablename__ = "state"
    state_id: Mapped[str] = mapped_column(unique=True, primary_key=True)

    updated_at: Mapped[t.Optional[datetime]] = mapped_column(  # noqa: UP007
        onupdate=datetime.now,
    )

    partial_state: Mapped[t.Optional[StateType]]  # noqa: UP007
    completed_state: Mapped[t.Optional[StateType]]  # noqa: UP007

    def __eq__(self, other: object) -> bool:
        """Check equality with another JobState.

        Does not take into account updated_at;
        merely checks equality of state_id and partial and completed state.

        Args:
            other: the other JobState to check against

        Returns:
            True if other is equal to this JobState, else False
        """
        if not isinstance(other, JobState):
            return NotImplemented
        return (
            (self.state_id == other.state_id)
            and (self.partial_state == other.partial_state)
            and (self.completed_state == other.completed_state)
        )

    @classmethod  # noqa: WPS210
    def from_job_history(cls, session: Session, state_id: str):  # noqa: WPS210
        """Build JobState from job run history.

        Args:
            session: the session to use in finding job history
            state_id: state_id to build JobState for

        Returns:
            JobState built from job run history
        """
        completed_state: dict[t.Any, t.Any] = {}
        partial_state: dict[t.Any, t.Any] = {}
        incomplete_since = None
        finder = JobFinder(state_id)

        # Get the state for the most recent completed job.
        # Do not consider dummy jobs create via add_state.
        if state_job := finder.latest_with_payload(session, flags=Payload.STATE):
            incomplete_since = state_job.ended_at
            if SINGER_STATE_KEY in state_job.payload:
                merge(state_job.payload, partial_state)

        # If there have been any incomplete jobs since the most recent completed jobs,
        # merge the state emitted by those jobs into the state for the most recent
        # completed job. If there are no completed jobs, get the full history of
        # incomplete jobs and use the most recent state emitted per stream
        incomplete_state_jobs = finder.with_payload(
            session,
            flags=Payload.INCOMPLETE_STATE,
            since=incomplete_since,
        )
        for incomplete_state_job in incomplete_state_jobs:
            if SINGER_STATE_KEY in incomplete_state_job.payload:
                partial_state = merge(incomplete_state_job.payload, partial_state)

        return cls(
            state_id=state_id,
            partial_state=partial_state,
            completed_state=completed_state,
        )

    @classmethod
    def from_json(cls, state_id: str, json_str: str) -> JobState:
        """Create JobState from json representation.

        Args:
            state_id: the state ID of the job to create JobState for.
            json_str: the json representation of state to use.

        Returns:
            JobState representing args.
        """
        state_dict = json.loads(json_str)
        return cls(
            state_id=state_id,
            completed_state=state_dict.get("completed", {}),
            partial_state=state_dict.get("partial", {}),
        )

    def json(self) -> str:
        """Get the json representation of this JobState.

        Returns:
            json representation of this JobState
        """
        return json.dumps(
            {"completed": self.completed_state, "partial": self.partial_state},
        )

    def json_merged(self) -> str:
        """Return the json representation of partial state merged onto complete state.

        Returns:
            json representation of partial state merged onto complete state.
        """
        return json.dumps(merge(self.partial_state, self.completed_state))

    @classmethod
    def from_file(cls, state_id: str, file_obj: TextIOWrapper) -> JobState:
        """Create JobState from a file-like object containing state json.

        Args:
            state_id: the state_id for the JobState
            file_obj: the file-like object containing state json.

        Returns:
            JobState
        """
        return cls.from_json(state_id=state_id, json_str=file_obj.read())

    def merge_partial(
        self,
        state: JobState,
    ):
        """Merge provided partial state onto this JobState.

        Args:
            state: the partial state to merge onto this state.
        """
        self.partial_state = merge(  # noqa: WPS601
            state.partial_state,
            self.partial_state,
        )

    def is_complete(self) -> bool:
        """Check if this JobState is complete.

        Returns:
            True if complete, else False
        """
        return bool(self.completed_state)

    def to_file(self, file_obj: TextIOWrapper):
        """Persist JobState to a file-like object.

        Args:
            file_obj: the file-like object to write to.
        """
        file_obj.write(self.json())



from __future__ import annotations

import platform
import threading
import time
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool

import pytest

from meltano.core.behavior.versioned import IncompatibleVersionError
from meltano.core.error import ProjectNotFound
from meltano.core.project import PROJECT_ROOT_ENV, Project


@pytest.fixture()
def deactivate_project(project):
    Project.deactivate()
    yield
    Project.activate(project)


def update(payload):
    project = Project.find()

    with project.meltano_update() as meltano:
        for k, v in payload.items():
            setattr(meltano, k, v)


class IndefiniteThread(threading.Thread):
    """Never ending thread."""

    def __init__(self):
        """Set stop event."""
        super().__init__()
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def run(self, *args, **kwargs):
        while not self._stop_event.is_set():
            self.do(*args, **kwargs)


class ProjectReader(IndefiniteThread):
    """Project using a never ending thread."""

    def __init__(self, project):
        """Set the project."""
        self.project = project
        super().__init__()

    def do(self):
        assert self.project.meltano
        time.sleep(50 / 1000)  # 50ms


class TestProject:
    @pytest.mark.usefixtures("deactivate_project")
    def test_find(self, project, tmp_path, monkeypatch):
        # defaults to the cwd
        found = Project.find(activate=False)
        assert found == project

        # or you can specify a path
        found = Project.find(project.root, activate=False)
        assert found == project

        # or set the MELTANO_PROJECT_ROOT env var
        with monkeypatch.context() as ctx1:
            ctx1.chdir(project.root.joinpath("extract"))
            ctx1.setenv(PROJECT_ROOT_ENV, "..")

            found = Project.find(activate=False)
            assert found == project

        # it can also recurse up from a subdirectory
        with monkeypatch.context() as ctx2:
            ctx2.chdir(project.root.joinpath("extract"))
            found = Project.find(activate=False)
            assert found == project

        # and it fails if there isn't a meltano.yml
        with pytest.raises(ProjectNotFound):
            Project.find(tmp_path)

    def test_activate(self, project):
        Project.deactivate()
        assert Project._default is None

        Project.activate(project)

        assert Project._default is project
        assert Project.find() is project

    def test_find_threadsafe(self, project, concurrency):
        workers = ThreadPool(concurrency["threads"])
        projects = workers.map(Project.find, range(concurrency["cases"]))
        assert all(x is project for x in projects)

    @pytest.mark.concurrent()
    def test_meltano_concurrency(self, project, concurrency):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        payloads = [{f"test_{i}": i} for i in range(1, concurrency["cases"] + 1)]

        reader = ProjectReader(project)
        reader.start()

        workers = Pool(concurrency["processes"])
        workers.map(update, payloads)

        reader.stop()
        reader.join()

        meltano = project.meltano
        unpacked_items = (item for payload in payloads for item in payload.items())
        for key, val in unpacked_items:
            assert meltano.extras[key] == val

    def test_preserve_comments(self, project: Project):
        original_contents = project.meltanofile.read_text()

        new_contents = f"# Please don't delete me :)\n{original_contents}"
        project.meltanofile.write_text(new_contents)

        with project.meltano_update() as meltano:
            meltano.extras["a_new_key"] = "New Key"

        contents = project.meltanofile.read_text()
        assert contents.startswith("# Please don't delete me :)\n")
        assert "a_new_key: New Key" in contents


class TestIncompatibleProject:
    def test_incompatible(self, project):
        with project.meltano_update() as meltano:
            meltano["version"] += 1

        with pytest.raises(IncompatibleVersionError):
            Project.activate(project)



from __future__ import annotations

import datetime
import json
import os
import platform
import shutil
import string
import typing as t
from base64 import b64encode
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob._models import BlobProperties  # noqa: WPS436
from boto3 import client
from botocore.stub import Stubber
from dateutil.tz import tzutc
from google.cloud.storage import Blob, Bucket

from meltano.core.job_state import JobState
from meltano.core.state_store import (
    AZStorageStateStoreManager,
    GCSStateStoreManager,
    LocalFilesystemStateStoreManager,
    S3StateStoreManager,
    WindowsFilesystemStateStoreManager,
)

if t.TYPE_CHECKING:
    from collections.abc import Iterator


def on_windows() -> bool:
    return "Windows" in platform.system()


def encode_if_on_windows(string: str) -> str:
    if on_windows():
        return b64encode(string.encode()).decode()
    return string


class TestLocalFilesystemStateStoreManager:
    @pytest.fixture()
    def subject(self, function_scoped_test_dir):
        if on_windows():
            yield WindowsFilesystemStateStoreManager(
                uri=f"file://{function_scoped_test_dir}\\.meltano\\state\\",
                lock_timeout_seconds=10,
            )
        else:
            yield LocalFilesystemStateStoreManager(
                uri=f"file://{function_scoped_test_dir}/.meltano/state/",
                lock_timeout_seconds=10,
            )

    @pytest.fixture()
    def state_path(
        self,
        function_scoped_test_dir,
        subject: LocalFilesystemStateStoreManager,
    ):
        Path(subject.state_dir).mkdir(parents=True, exist_ok=True)
        yield os.path.join(function_scoped_test_dir, ".meltano", "state")
        shutil.rmtree(
            os.path.join(function_scoped_test_dir, ".meltano", "state"),
            ignore_errors=True,
        )

    def test_join_path(self, subject: LocalFilesystemStateStoreManager):
        if on_windows():
            assert subject.join_path("a", "b") == "a\\b"
            assert subject.join_path("a", "b", "c", "d", "e") == "a\\b\\c\\d\\e"
        else:
            assert subject.join_path("a", "b") == "a/b"
            assert subject.join_path("a", "b", "c", "d", "e") == "a/b/c/d/e"

    def test_create_state_id_dir_if_not_exists(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
    ):
        state_id_path = os.path.join(
            state_path,
            encode_if_on_windows("create_state_id_dir"),
        )
        assert not os.path.exists(state_id_path)
        subject.create_state_id_dir_if_not_exists("create_state_id_dir")
        assert os.path.exists(state_id_path)

    def test_get_reader(self, subject: LocalFilesystemStateStoreManager, state_path):
        filepath = os.path.join(state_path, "get_reader")
        open(filepath, "a").close()  # noqa: WPS515, SIM115
        with subject.get_reader(path=filepath) as reader:
            assert reader.name == filepath

    def test_get_writer(self, subject: LocalFilesystemStateStoreManager, state_path):
        filepath = os.path.join(state_path, "get_writer")
        with subject.get_writer(path=filepath) as writer:
            assert writer.name == filepath

    def test_get_state_path(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
    ):
        assert subject.get_state_path("get_state_path") == os.path.join(
            state_path,
            encode_if_on_windows("get_state_path"),
            "state.json",
        )

    def test_get_lock_path(self, subject: LocalFilesystemStateStoreManager, state_path):
        assert subject.get_lock_path("some_state_id") == os.path.join(
            state_path,
            encode_if_on_windows("some_state_id"),
            "lock",
        )

    def test_acquire_lock(self, subject: LocalFilesystemStateStoreManager, state_path):
        dir_path = os.path.join(state_path, encode_if_on_windows("acquire_lock"))
        with subject.acquire_lock("acquire_lock"):
            assert os.path.exists(os.path.join(dir_path, "lock"))

    @pytest.mark.usefixtures("state_path")
    def test_get_state_ids(self, subject: LocalFilesystemStateStoreManager):
        dev_ids = [f"dev:{letter}-to-{letter}" for letter in string.ascii_lowercase]
        prod_ids = [f"prod:{letter}-to-{letter}" for letter in string.ascii_lowercase]
        for state_id in dev_ids + prod_ids:
            Path(subject.get_path(state_id)).mkdir(parents=True)
            open(  # noqa: WPS515, SIM115
                subject.get_path(state_id, filename="state.json"),
                "w+",
            ).close()
        assert set(dev_ids + prod_ids) == set(subject.get_state_ids())
        assert set(dev_ids) == set(subject.get_state_ids(pattern="dev*"))

    def test_get(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
        state_ids_with_expected_states,
    ):
        for state_id, expected_state in state_ids_with_expected_states:
            state_dir = os.path.join(state_path, encode_if_on_windows(state_id))
            Path(state_dir).mkdir(parents=True)
            with open(  # noqa: WPS515
                os.path.join(state_dir, "state.json"),
                "w+",
            ) as state_file:
                json.dump(expected_state, state_file)

        for state_id, expected_state in state_ids_with_expected_states:
            assert subject.get(state_id) == JobState.from_json(
                state_id,
                json.dumps(expected_state),
            )

    def test_set(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
        state_ids_with_expected_states,
    ):
        for state_id, expected_state in state_ids_with_expected_states:
            subject.set(
                JobState.from_json(state_id, json.dumps({"completed": expected_state})),
            )
        for state_id, expected_state in state_ids_with_expected_states:
            with open(
                os.path.join(state_path, encode_if_on_windows(state_id), "state.json"),
            ) as state_file:
                assert JobState.from_json(
                    state_id,
                    json.dumps({"completed": expected_state}),
                ) == JobState.from_file(state_id, state_file)

    def test_delete(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
        state_ids_with_expected_states,
    ):
        # Delete files
        state_id, expected_state = state_ids_with_expected_states[0]
        state_dir = os.path.join(state_path, encode_if_on_windows(state_id))
        Path(state_dir).mkdir(parents=True)
        filepath = os.path.join(state_dir, "state.json")
        with open(filepath, "w+") as state_file:
            json.dump(expected_state, state_file)
        assert os.path.exists(filepath)
        subject.delete(filepath)
        assert not os.path.exists(filepath)

        # Delete directories
        assert os.path.exists(state_dir)
        subject.delete(state_dir)
        assert not os.path.exists(state_dir)

        # Swallows FileNotFoundError
        subject.delete(filepath)
        subject.delete(state_dir)

    def test_clear(
        self,
        subject: LocalFilesystemStateStoreManager,
        state_path,
        state_ids_with_expected_states,
    ):
        for state_id, expected_state in state_ids_with_expected_states:
            state_dir = os.path.join(state_path, encode_if_on_windows(state_id))
            Path(state_dir).mkdir(parents=True)
            with open(os.path.join(state_dir, "state.json"), "w+") as state_file:
                json.dump(expected_state, state_file)
        for state_id, _ in state_ids_with_expected_states:
            state_dir = os.path.join(state_path, encode_if_on_windows(state_id))
            filepath = os.path.join(state_dir, "state.json")
            assert os.path.exists(filepath)
            subject.clear(state_id)
            assert not os.path.exists(os.path.dirname(filepath))


class TestAZStorageStateStoreManager:
    @pytest.fixture()
    def subject(
        self,
        function_scoped_test_dir,  # noqa: ARG002
    ):
        return AZStorageStateStoreManager(
            uri="azure://meltano/state/",
            connection_string="UseDevelopmentStorage=true",
            lock_timeout_seconds=10,
        )

    @pytest.fixture()
    def mock_client(self):
        with patch(
            "meltano.core.state_store.azure.BlobServiceClient",
        ) as mock_client:
            yield mock_client

    def test_client(self, subject: AZStorageStateStoreManager, mock_client):
        # Call twice to assure memoization
        _ = subject.client  # noqa: F541 WPS122
        _ = subject.client  # noqa: F541 WPS122
        mock_client.from_connection_string.assert_called_once_with(
            "UseDevelopmentStorage=true",
        )

    @pytest.mark.usefixtures("mock_client")
    def test_is_file_not_found_error_true(
        self,
        subject: AZStorageStateStoreManager,
    ):
        got_reader = False
        mock_container_client = MagicMock()
        mock_container_client.container_name = subject.container_name
        subject.client.get_container_client.return_value = mock_container_client
        mock_container_client.get_blob_client.side_effect = ResourceNotFoundError(
            "Operation returned an invalid status 'The specified blob does "
            "not exist.'\nErrorCode:BlobNotFound",
        )
        try:
            with subject.get_reader("nonexistent"):
                got_reader = True
        except Exception as e:
            assert subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    @pytest.mark.usefixtures("mock_client")
    def test_is_file_not_found_error_false(
        self,
        subject: AZStorageStateStoreManager,
    ):
        got_reader = False
        mock_container_client = MagicMock()
        mock_container_client.container_name = subject.container_name
        subject.client.get_container_client.return_value = mock_container_client
        mock_container_client.get_blob_client.side_effect = ResourceNotFoundError(
            "Operation returned an invalid status 'The specified container "
            "does not exist.'\nErrorCode:ContainerNotFound",
        )
        try:
            with subject.get_reader("nonexistent"):
                got_reader = True
        except Exception as e:
            assert not subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    def test_state_path(self, subject: AZStorageStateStoreManager):
        assert subject.state_dir == "state"

    @pytest.mark.usefixtures("mock_client")
    def test_delete(self, subject):
        mock_blob_client = MagicMock()
        subject.client.get_blob_client.return_value = mock_blob_client
        subject.delete("some_path")
        mock_blob_client.delete_blob.assert_called_once()

    @pytest.mark.usefixtures("mock_client")
    def test_get_state_ids(self, subject):
        mock_container_client = MagicMock()
        mock_container_client.list_blobs.return_value = (
            BlobProperties(name=f"state/state_id_{i}/state.json") for i in range(10)
        )
        subject.client.get_container_client.return_value = mock_container_client
        assert set(subject.get_state_ids()) == {f"state_id_{i}" for i in range(10)}
        mock_container_client.list_blobs.assert_called_once_with(
            name_starts_with="state/",
        )


class TestS3StateStoreManager:
    @contextmanager
    def stubber(self) -> Iterator[Stubber]:
        with patch(
            "meltano.core.state_store.S3StateStoreManager.client",
            new_callable=PropertyMock,
        ) as mock_client:
            mock_client.return_value = client("s3")
            with Stubber(mock_client.return_value) as stubber:
                yield stubber

    @pytest.fixture()
    def subject(
        self,
        function_scoped_test_dir,  # noqa: ARG002
    ):
        return S3StateStoreManager(
            uri="s3://test_access_key_id:test_secret_access_key@meltano/state",
            container_name="testing",
            lock_timeout_seconds=10,
        )

    def test_is_file_not_found_error_true(self, subject: S3StateStoreManager):
        got_reader = False
        with self.stubber() as stubber:
            stubber.add_client_error("get_object", service_error_code="NoSuchKey")
            try:
                with subject.get_reader("does_not_exist"):
                    got_reader = True
            except Exception as e:
                assert subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    def test_is_file_not_found_error_false(self, subject: S3StateStoreManager):
        got_reader = False
        with self.stubber() as stubber:
            stubber.add_client_error("get_object", service_error_code="NoSuchBucket")
            try:
                with subject.get_reader("does_not_exist"):
                    got_reader = True
            except Exception as e:
                assert not subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    def test_client_session(self, subject: S3StateStoreManager):
        with patch("boto3.Session") as mock_session:
            _ = subject.client  # noqa: F541 WPS122
            _ = subject.client  # noqa: F541 WPS122
            mock_session.assert_called_once_with(
                aws_access_key_id=subject.aws_access_key_id,
                aws_secret_access_key=subject.aws_secret_access_key,
            )

    def test_client_client(self, subject: S3StateStoreManager):
        with patch("boto3.Session.client") as mock_client:
            _ = subject.client  # noqa: F541 WPS122
            _ = subject.client  # noqa: F541 WPS122
            mock_client.assert_called_once_with("s3", endpoint_url=subject.endpoint_url)

    def test_state_path(self, subject: S3StateStoreManager):
        assert subject.state_dir == "state"

    def test_delete(self, subject: S3StateStoreManager):
        response = {
            "ResponseMetadata": {
                "RequestId": "test_delete",
                "HostId": "",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "accept-ranges": "bytes",
                    "content-length": "236",
                    "content-security-policy": "block-all-mixed-content",
                    "content-type": "application/xml",
                    "server": "test_delete",
                    "strict-transport-security": "max-age=31536000; includeSubDomains",
                    "vary": "Origin, Accept-Encoding",
                    "x-amz-request-id": "test_delete",
                    "x-content-type-options": "nosniff",
                    "x-xss-protection": "1; mode=block",
                    "date": "Mon, 24 Oct 2022 01:05:07 GMT",
                },
                "RetryAttempts": 0,
            },
            "Deleted": [
                {"Key": "/state/test_delete"},
            ],
        }
        with self.stubber() as stubber:
            stubber.add_response(
                "delete_objects",
                response,
                expected_params={
                    "Bucket": subject.bucket,
                    "Delete": {"Objects": [{"Key": "/state/test_delete"}]},
                },
            )
            subject.delete("/state/test_delete")

    def test_get_state_ids(self, subject: S3StateStoreManager):
        response = {
            "ResponseMetadata": {
                "RequestId": "test_get_state_ids",
                "HostId": "",
                "HTTPStatusCode": 200,
                "HTTPHeaders": {
                    "accept-ranges": "bytes",
                    "content-length": "993",
                    "content-security-policy": "block-all-mixed-content",
                    "content-type": "application/xml",
                    "server": "test_get_state_ids",
                    "strict-transport-security": "max-age=31536000; includeSubDomains",
                    "vary": "Origin, Accept-Encoding",
                    "x-amz-request-id": "test_get_state_ids",
                    "x-content-type-options": "nosniff",
                    "x-xss-protection": "1; mode=block",
                    "date": "Wed, 26 Oct 2022 01:15:33 GMT",
                },
                "RetryAttempts": 0,
            },
            "IsTruncated": False,
            "Contents": [
                {
                    "Key": "state/state_id_2/state.json",
                    "LastModified": datetime.datetime(
                        2022,
                        10,
                        26,
                        1,
                        15,
                        26,
                        173000,
                        tzinfo=tzutc(),
                    ),
                    "ETag": '"test_get_state_ids"',
                    "Size": 60,
                    "StorageClass": "STANDARD",
                    "Owner": {
                        "DisplayName": "test_get_state_ids",
                        "ID": "test_get_state_ids",
                    },
                },
                {
                    "Key": "state/state_id_1/state.json",
                    "LastModified": datetime.datetime(
                        2022,
                        10,
                        24,
                        1,
                        15,
                        12,
                        450000,
                        tzinfo=tzutc(),
                    ),
                    "ETag": '"test_get_state_ids"',
                    "Size": 60,
                    "StorageClass": "STANDARD",
                    "Owner": {
                        "DisplayName": "test_get_state_ids",
                        "ID": "test_get_state_ids",
                    },
                },
            ],
            "Name": subject.bucket,
            "Delimiter": "",
            "MaxKeys": 1000,
            "EncodingType": "url",
            "KeyCount": 2,
        }
        with self.stubber() as stubber:
            stubber.add_response(
                "list_objects_v2",
                response,
                expected_params={"Bucket": subject.bucket},
            )
            assert set(subject.get_state_ids()) == {"state_id_1", "state_id_2"}


class TestGCSStateStoreManager:
    @pytest.fixture()
    def subject(
        self,
        function_scoped_test_dir,  # noqa: ARG002
    ):
        return GCSStateStoreManager(
            uri="gs://meltano/state/",
            application_credentials="path/to/creds/file",
            lock_timeout_seconds=10,
        )

    @pytest.fixture()
    def mock_client(self):
        with patch(
            "google.cloud.storage.Client",
        ) as mock_client:
            yield mock_client

    def test_client(self, subject: GCSStateStoreManager, mock_client):
        # Call twice to assure memoization
        _ = subject.client  # noqa: F541 WPS122
        _ = subject.client  # noqa: F541 WPS122
        mock_client.from_service_account_json.assert_called_once_with(
            subject.application_credentials,
        )

    @pytest.mark.usefixtures("mock_client")
    def test_is_file_not_found_error_true(
        self,
        subject: GCSStateStoreManager,
    ):
        got_reader = False
        mock_bucket = MagicMock()
        mock_bucket.get_blob.return_value = None
        subject.client.bucket.return_value = mock_bucket
        try:
            with subject.get_reader("nonexistent"):
                got_reader = True
        except Exception as e:
            assert subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    @pytest.mark.usefixtures("mock_client")
    def test_is_file_not_found_error_false(
        self,
        subject: GCSStateStoreManager,
    ):
        got_reader = False
        mock_blob = MagicMock()
        mock_blob.open.side_effect = ValueError("Some other error")
        mock_bucket = MagicMock()
        mock_bucket.blob.return_value = mock_blob
        subject.client.bucket.return_value = mock_bucket
        try:
            with subject.get_reader("nonexistent"):
                got_reader = True
        except Exception as e:
            assert not subject.is_file_not_found_error(e)  # noqa: PT017
        assert not got_reader

    def test_state_path(self, subject: GCSStateStoreManager):
        assert subject.state_dir == "state"

    @pytest.mark.usefixtures("mock_client")
    def test_delete(self, subject: GCSStateStoreManager):
        mock_blob = MagicMock()
        mock_bucket = MagicMock()
        mock_bucket.blob.return_value = mock_blob
        subject.client.bucket.return_value = mock_bucket
        subject.delete("some_path")
        mock_blob.delete.assert_called_once()

    @pytest.mark.usefixtures("mock_client")
    def test_get_state_ids(self, subject: GCSStateStoreManager):
        subject.client.list_blobs.return_value = (
            Blob(bucket=Bucket("meltano"), name=f"state/state_id_{i}/state.json")
            for i in range(10)
        )
        assert set(subject.get_state_ids()) == {f"state_id_{i}" for i in range(10)}
        subject.client.list_blobs.assert_called_once_with(bucket_or_name="meltano")
